from .sobol import saltelli_with_constant_bounds, get_indices
from .cluster_based import get_time_series_dataset, get_kmeans_clusters

__all__ = ["saltelli_with_constant_bounds",
           "get_indices",
           "get_time_series_dataset",
           "get_kmeans_clusters"]



import pandas as pd


def get_mean(df_in: pd.DataFrame, time_name: str, quantity_name: str) -> pd.DataFrame:
    """Obtain the mean time series from a pandas dataframe of timeseries

    Args:
        df_in: input dataframe with all time series. Columns: "run_name","time_name","quantity_name",...
        time_name: Name of column containing time values.
        quantity_name: Name of column containing the model run indices.

    Returns:
        df_mean: Dataframe containing the mean time series
    """
    df_mean = df_in.groupby(time_name, as_index=False)[quantity_name].mean()
    return df_mean



import pandas as pd
import ujson


class SensitivityInput:
    def __init__(self, data: dict, parameters: list, bounds: list, quantity_mean: str, quantity_variance: str):
        """Class for sensitivity analysis input data

        Args:
            data: Parameters and outputs for multiple model runs in JSON format.
                The columns are columns: "self.parameters, self.quantity_mean, self.quantity_variance"
            parameters: Names of model parameters, which can include constants, as ["param_1","param_2","param_x",...]
            bounds: Bounds of model parameters as [[param1_lower_bound,param1_upper_bound],[...],[param_x_fixed],..].
                For parameters with a fixed value provide one entry with the fixed value.
            quantity_mean: Name of the column containing the mean of the output quantity.
            quantity_variance: Name of the column containing the variance of the output quantity.
        """
        self.parameters = parameters
        self.bounds = bounds
        self.quantity_mean = quantity_mean
        self.quantity_variance = quantity_variance
        self.data = data

    def df(self) -> pd.DataFrame:
        """"Returns self.data as a pandas dataframe.
        If the "quantity_mean" and "quantity  variance" entries are strings containing lists, ie. "[1,2,3,...]" then
        they are converted into python lists and added to the dataframe.
        Thus, time series when stored as lists in the dataframe's output columns are handled correctly.
        """
        df = pd.read_json(self.data, orient='split')

        if df[self.quantity_mean].map(type).eq(str).all() and df[self.quantity_variance].map(type).eq(str).all():
            try:
                temp_quantity_mean = df[self.quantity_mean].map(lambda x: ujson.loads(x), na_action='ignore')
                temp_quantity_variance = df[self.quantity_variance].map(lambda x: ujson.loads(x), na_action='ignore')
                df[self.quantity_mean] = temp_quantity_mean
                df[self.quantity_variance] = temp_quantity_variance
            except ValueError:
                print("\"quantity_mean\", \"quantity_variance\" entries not converted: strings but not in json format")
        return df



import pandas as pd


class UncertaintyInput:
    def __init__(self, data: dict, quantity_name: str, time_name: str, run_name: str):
        """Class for uncertainty quantification input data.

        Args:
            data: Parameters and outputs for multiple model runs in JSON format (orient='split').
                The columns are: "run_name","time_name","quantity_name",...
            quantity_name: Name of column containing the output quantitiy of interest
            time_name: Name of column containing time values.
            run_name: Name of column containing the model run indices.
        """
        self.data = data
        self.quantity_name = quantity_name
        self.time_name = time_name
        self.run_name = run_name

    def df(self) -> pd.DataFrame:
        """"Returns self.data as a pandas dataframe.

        This is a method so it is not included in the __dict__ of UncertaintyInput objects.
        """
        return pd.read_json(self.data, orient='split')



import numpy as np
import pandas as pd
import sklearn.gaussian_process as gp
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
from typing import Tuple, Optional, Callable
import random


def split_training_test_set(df_in: pd.DataFrame, test_set_size_in: int, seed: int = 42) -> Tuple[
    pd.DataFrame, pd.DataFrame]:
    """

    Splits a Dataframe into two, by randomly sampling rows, so one can be used for training and the other for testing.

    Args:
        df_in: Dataframe of model data, each row is a different sample.
        Columns are, parameters, output mean, output variance.
        test_set_size_in: Number of samples to move to a test set.
        seed: Random seed for drawing the test set.  Default: 42.

    Returns:
        df_training: Subset of df_in chosen as training set.
        df_test: Subset of df_in chosen as test set.
    """

    total_runs = len(df_in.index)
    test_set_size = test_set_size_in
    if total_runs <= test_set_size_in:
        raise ValueError("All of the data (or more) is assigned to training set, this is not sensible.")

    random.seed(seed)
    testset = random.sample(range(total_runs), test_set_size)
    testset.sort()
    df_training = df_in.drop(testset)
    df_test = df_in.iloc[testset]
    return df_training, df_test


def train_GP_emulator(X_in: np.ndarray, y_in: np.ndarray, alpha_in: np.ndarray) -> Tuple[
    gp.GaussianProcessRegressor, StandardScaler]:
    """

    Returns a model trained on scaled inputs, also returns the associated Scaler object.

    Args:
        X_in: Model parameter values, need not be normalised.
        y_in: Model output values, need not be normalised (they would never need to anyway).
        alpha_in: The variance of the model outputs. From SKLearn Docs:
            https://scikit-learn.org/stable/modules/generated/sklearn.gaussian_process.GaussianProcessRegressor.html
            " It can be interpreted as the variance of additional Gaussian measurement noise on the training observations".

    Returns:
        gp_model: Trained Gaussian model object.
        gp_scaler: Scaler object for input data.
    """

    # Gaussian process section
    # Scaling input parameters
    gp_scaler = StandardScaler()
    gp_scaler.fit(X_in)

    kernel = gp.kernels.ConstantKernel(1.0, [1e-7, 1e15]) * gp.kernels.RBF(10.0)
    gp_model = gp.GaussianProcessRegressor(kernel=kernel, n_restarts_optimizer=100, alpha=alpha_in)
    gp_model.fit(gp_scaler.transform(X_in), y_in)

    return gp_model, gp_scaler


def predict_GP_emulator(X_in: np.ndarray, model_in: gp.GaussianProcessRegressor, scaler_in: StandardScaler,
                        return_std_in: bool = False) -> Tuple[np.ndarray, Optional[np.ndarray]]:
    """

    Makes predictions given parameter values, a model and a Scaler.

    Args:
        X_in: Parameters to predict output for, need not be normalised.
        model_in: Model to be evaluated, preferably a trained model.
        scaler_in: The Scaler associated with the trained model.
        return_std_in: Whether or not to return the standard deviation of the predictions

    Returns:
        y_pred: Prediction of output for parameters given by X_in.
        std_out: standard deviation associated with y_pred_out (if return_std_in = True).
    """

    return model_in.predict(scaler_in.transform(X_in), return_std=return_std_in)


def evaluate_GP_emulator(X_in: np.ndarray, y_in: np.ndarray, model_in: gp.GaussianProcessRegressor,
                         scaler_in: StandardScaler):
    """

    Evaluate the Gaussian process emulator on a test set and plot the results.
    Do not use the set you used for training...

    Args:
        X_in: Test set parameters, need not be normalised.
        y_in: Correct output for training examples.
        model_in: Model to be evaluated.
        scaler_in: The Scaler associated with the model.
    """
    # Predict labels
    y_pred, std = predict_GP_emulator(X_in, model_in, scaler_in, True)

    # Plot validation figure
    plt.figure(figsize=(10.5, 6))
    plt.plot(range(len(y_in)), y_in, 'o', color='black', label='True Value')
    plt.plot(range(len(y_pred)), y_pred, 'o', color='red', label='Prediction')
    plt.plot(range(len(y_in)), y_pred - 1.9600 * std, '--', color='black')
    plt.plot(range(len(y_in)), y_pred + 1.9600 * std, '--', color='black', label='95% confidence interval')
    plt.ylabel('Total Hospital Deaths Over 200 days', fontsize=15)
    plt.xlabel('Test points', fontsize=15)
    plt.ylim(0)
    plt.title('Gaussian Process Emulator Test Set Performance', fontsize=15)
    plt.legend(loc='lower left')
    plt.show()
    return


def train_and_predict(df_in: pd.DataFrame, params_in: list, quantity_mean_in: str, quantity_variance_in: str,
                      X_in: np.ndarray, scalar_mean_function: Callable[[list], float],
                      scalar_variance_function: Callable[[list], float]) -> np.ndarray:
    """

    Takes training data and parameter values to be evaluated and returns the GP emulator predictions at those values.

    Args:
        df_in: Dataframe with input parameters and the mean and variance of the output quantity.
        params_in: Names of the columns with model parameters in df_in.
        quantity_mean_in: name of the column with the mean of the model output to be analysed.
        quantity_variance_in: name of the column with the variance of the model output to be analysed.
        X_in: Numpy array with the parameters to be tested, each row being a different set of parameters.
        scalar_mean_function: Function mapping list objects in the mean column of df_in to scalars.
        scalar_variance_function: Function mapping list objects in the variance column of df_in to scalars.

    Returns:
        y_out: Gaussian process estimator predictions for each set of parameters in X_in.
    """
    df = get_scalar_features(df_in, quantity_mean_in, quantity_variance_in, scalar_mean_function,
                             scalar_variance_function)
    X_tr_temp, y_tr_temp, alpha_tr_temp = form_training_set(df, params_in, quantity_mean_in, quantity_variance_in)
    temp_model, temp_scaler = train_GP_emulator(X_tr_temp, y_tr_temp, alpha_tr_temp)
    y_out = predict_GP_emulator(X_in, temp_model, temp_scaler)
    return y_out


def get_scalar_features(df_in: pd.DataFrame, quantity_mean_in: list, quantity_variance_in: list,
                        scalar_mean_function: Callable[[list], float],
                        scalar_variance_function: Callable[[list], float]) -> pd.DataFrame:
    """Applies functions mapping model outputs from lists to scalars.
        Since the gaussian process emulator trains on scalar outputs.

    Args:
        df_in: Dataframe of model data, each row is a different sample.
            Columns are: parameters, output mean, output variance.
        quantity_mean_in: Name of the column containing the mean of the output quantity.
        quantity_variance_in: Name of the column containing the variance of the output quantity.
        scalar_mean_function: Function mapping list objects in the mean column of df_in to scalars.
        scalar_variance_function: Function mapping list objects in the variance column of df_in to scalars.

    Returns:
        df: Dataframe with scalar entries in the mean and variance columns, given by scalar_mean/variance_functions.
    """
    # Check if columns contain lists
    if df_in[quantity_mean_in].map(type).eq(list).all() and df_in[quantity_variance_in].map(type).eq(list).all():
        # Apply functions mapping lists to scalars to each column entry
        df_in[quantity_mean_in] = df_in[quantity_mean_in].apply(scalar_mean_function)
        df_in[quantity_variance_in] = df_in[quantity_variance_in].apply(scalar_variance_function)
    return df_in


def form_training_set(df_in: pd.DataFrame, params_in: list, quantity_mean_in: str, quantity_variance_in: str) -> Tuple[
    np.ndarray, np.ndarray, np.ndarray]:
    """
    Forms numpy arrays from a dataframe, the numpy arrays can then be used for training a model.

    Args:
        df_in: Dataframe with input parameters and the mean and variance of the output quantity.
        params_in: Names of the columns with model parameters in df_in.
        quantity_mean_in: Name of the column with the mean of the model output to be analysed.
        quantity_variance_in: Name of the column with the variance of the model output to be analysed.

    Returns:
        X_out: Numpy array with training set parameters, used for training a model.
        y_out: Numpy array with training set outputs, used for training a model.
        alpha_out: Numpy array with the variance of the training set outputs, can be used for training a model.
    """

    # Training set
    params_temp = df_in[params_in]
    X_out = params_temp.to_numpy()  # training parameters

    output_mean_temp = df_in[quantity_mean_in]
    y_out = output_mean_temp.to_numpy()  # training outputs

    output_variance_temp = df_in[quantity_variance_in]
    alpha_out = output_variance_temp.to_numpy()  # variance at training points, as described in sklearn docs
    return X_out, y_out, alpha_out


def form_test_set(df_in: pd.DataFrame, params_in: list, quantity_mean_in: str) -> Tuple[np.ndarray, np.ndarray]:
    """

    Forms numpy arrays from a dataframe, the numpy arrays can be used for testing and evaluating a model.

    Args:
        df_in: Dataframe with input parameters and the mean and variance of the output quantity.
        params_in: Names of the columns with model parameters in df_in.
        quantity_mean_in: Name of the column with the mean of the model output.

    Returns:
        X_out: Numpy array with test set parameters.
        y_out: Numpy array with test set outputs.

    """
    params_temp = df_in[params_in]  # test set parameters
    X_out = params_temp.to_numpy()

    output_mean_temp = df_in[quantity_mean_in]  # correct test set labels
    y_out = output_mean_temp.to_numpy()
    return X_out, y_out



import numpy as np
import pandas as pd
from tslearn.clustering import TimeSeriesKMeans
import collections


def get_time_series_dataset(df: pd.DataFrame, quantity_mean_in: str) -> np.ndarray:
    """Converts entries in quantity_mean column of the dataframe to a time series dataset for ts-learn.

    Args:
        df: Dataframe of model data, each row is a different sample.
            Columns are: parameters, output mean, output variance.
        quantity_mean_in: Name of the column containing the mean of the output quantity.

    Returns:
        time_series: Numpy array of the time series dataset format ts-learn takes as input.
    """

    nr_of_runs = df.shape[0]
    entry = df[quantity_mean_in][0]
    nr_of_time_samples = len(entry) if isinstance(entry, collections.Sized) else 1
    # make time series dataset
    time_series = np.zeros((nr_of_runs, nr_of_time_samples, 1))
    for idx in range(nr_of_runs):
        time_series[idx, :, 0] = df[quantity_mean_in][idx][:] if isinstance(entry, collections.Sized) else \
            df[quantity_mean_in][idx]
    return time_series


def get_kmeans_clusters(df: pd.DataFrame, quantity_mean_in: str, n_clusters: int, metric: str, verbose=False,
                        random_state=42) -> np.ndarray:
    """Computes cluster labels for time series stored in a column of a pandas dataframe, using kmeans clustering.

    Args:
        df: Dataframe of model data, each row is a different sample.
            Columns are: parameters, output mean, output variance.
        quantity_mean_in: Name of the column containing the mean of the output quantity.
        n_clusters: Number of clusters to find.
        metric: Metric to use, choices correspond to those for TimeSeriesKMeans, "euclidean", "dtw", "softdtw".
        verbose: If true, print information about inertia while learning.  Default: False
        random_state: Seed used to initialise cluster centres. Default: 42

    Returns:
        clusters: Numpy array containing the index of the cluster assigned to each parameter combination
        The order of the indices match the order of the rows in the dataframe.
    """
    time_series = get_time_series_dataset(df, quantity_mean_in)
    clusters = TimeSeriesKMeans(n_clusters=n_clusters, verbose=verbose, random_state=random_state,
                                metric=metric).fit_predict(time_series)
    return clusters



'''
Author: Erik Rydow
Email: erik.rydow@eng.ox.ac.uk
'''

from .gaussian_process_emulator import split_training_test_set, train_GP_emulator, predict_GP_emulator, \
    evaluate_GP_emulator, train_and_predict, form_training_set, form_test_set, get_scalar_features

from . import sensitivity_analysis
from . import uncertainty_quantification
from . import data_types

__all__ = ["split_training_test_set",
           "train_GP_emulator",
           "predict_GP_emulator",
           "evaluate_GP_emulator",
           "train_and_predict",
           "form_training_set",
           "form_test_set",
           "get_scalar_features",
           "sensitivity_analysis",
           "uncertainty_quantification",
           "data_types"
           ]



"""A setuptools based setup module.

See:
https://packaging.python.org/guides/distributing-packages-using-setuptools/
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

# Get the long description from the README file
long_description = (here / 'README.md').read_text(encoding='utf-8')

# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

setup(
    name='sandu',  # Required

    version='0.1.2',  # Required
    description='High Level Sensitivity and Uncertainty (SandU) Analysis Tools.',  # Optional

    long_description=long_description,  # Optional
    long_description_content_type='text/markdown',  # Optional (see note above)
    url='https://github.com/ErikRZH/sandu',  # Optional
    author='Erik Rydow',  # Optional
    author_email='erik.rydow@eng.ox.ac',  # Optional
    classifiers=[  # Optional

        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.8',

    ],
    keywords=['sensitivity-analysis', 'ensemble-analysis', 'uncertainty-analysis', 'uncertainty-quantification',
               'analysis-of-mathematical-models'],  # Optional

    # When your source code is in a subdirectory under the project root, e.g.
    # `src/`, it is necessary to specify the `package_dir` argument.
    package_dir={'': 'src'},  # Optional

    packages=find_packages(where='src'),  # Required

    python_requires='>=3.6, <4',

    install_requires=[
        'numpy',
        'pandas',
        'sklearn',
        'SALib',
        'matplotlib >= 3.5',
        'tslearn',
        'ujson'

    ],

    # If there are data files included in your packages that need to be
    # installed, specify them here.
    package_data={  # Optional
        'sample': ['package_data.dat'],
    },


    # List additional URLs that are relevant to your project as a dict.
    #
    # This field corresponds to the "Project-URL" metadata fields:
    # https://packaging.python.org/specifications/core-metadata/#project-url-multiple-use
    #
    # Examples listed include a pattern for specifying where the package tracks
    # issues, where the source is hosted, where to say thanks to the package
    # maintainers, and where to support the project financially. The key is
    # what's used to render the link text on PyPI.
    # project_urls={  # Optional
    #     'Bug Reports': 'https://github.com/pypa/sampleproject/issues',
    #     'Funding': 'https://donate.pypi.org',
    #     'Say Thanks!': 'http://saythanks.io/to/example',
    #     'Source': 'https://github.com/pypa/sampleproject/',
    # },
)


from SALib.analyze import sobol
from SALib.sample import saltelli
import pandas as pd
import numpy as np
from typing import Tuple, Callable

from .. import gaussian_process_emulator as gpe


def saltelli_with_constant_bounds(problem_in: dict, N_in: int) -> Tuple[np.ndarray, dict]:
    """

    Takes as input a problem dictionary, which may include constants(!), and an integer N.
    Returns a Saltelli sampling of the parameters, ignoring the constants, as a (D,N*(2D+2)) numpy array,
    and the problem dictionary with constants removed

    Arg:
        problem_in: Problem dictionary (keys: 'num_vars', 'names', 'bounds') which may contain constants.
        N_in: N used to determine number of parameter samples in Saltelli sampling.  Note: Must be power of 2.

    Returns:
        X_sample: Parameter values to be sampled in Sobol sensitivity analysis.
    """

    # Removes the constant parameters from the problem dictionary, generates samples and reinserts constant values
    constant_parameters = []
    for count, value in enumerate(problem_in['bounds']):
        if len(value) == 1:
            constant_parameters.append(count)
        elif len(value) != 2:
            raise ValueError("parameter named: " + problem_in['names'][
                count] + " has bounds with not 1 entry (constant) or 2 entries (bounds).")

    names_variables = [problem_in['names'][i] for i in range(len(problem_in['names'])) if
                       i not in constant_parameters]
    bounds_variables = [problem_in['bounds'][i] for i in range(len(problem_in['bounds'])) if
                        i not in constant_parameters]

    problem_variables = {'num_vars': len(names_variables),
                         'names': names_variables,
                         'bounds': bounds_variables}

    # Generates N*(2D+2) samples, where N is argument D is number of non-constant varaibles
    X_sample = saltelli.sample(problem_variables, N_in)

    # Reinsert Constant Values
    for i in constant_parameters:
        # Inserts columns with the constant value in correct place
        X_sample = np.insert(X_sample, i, np.ones([X_sample.shape[0]]) * problem_in['bounds'][i][0], axis=1)

    return X_sample, problem_variables


def get_indices(df_in: pd.DataFrame, params_in: list, bounds_in: list, quantity_mean_in: str,
                quantity_variance_in: str, N_in: int, scalar_mean_function: Callable[[list], float] = sum,
                scalar_variance_function: Callable[[list], float] = sum) -> pd.DataFrame:
    """

    Returns Sobol indices by sampling from a gaussian process emulator trained on model data.

    Args:
        df_in: Dataframe of model data, each row is a different sample.
            Columns are: parameters, output mean, output variance.
        params_in: Names of model parameters (can include constants).
        bounds_in: Bounds of model parameters as [[param1_lower_bound,param1_upper_bound],[...],[param_x_fixed],..].
            For parameters with a fixed value this is the fixed value.
        quantity_mean_in: Name of the column containing the mean of the output quantity.
        quantity_variance_in: Name of the column containing the variance of the output quantity.
        N_in: The number used in parameter sampling, should be a power of 2.
            N*(2D+2) samples are generated where D is the number of non-constant parameters.
        scalar_mean_function: Function mapping list objects in the mean column of df_in to scalars.
            Default: sum()
        scalar_variance_function: Function mapping list objects in the variance column of df_in to scalars.
            Default: sum()

    Returns:
        df_out: Dataframe containing 1st and total order sensitivity indices,
            and associated 95% confidence intervals.
    """
    # Set up problem dictionary,as needed by SALib
    problem_all = {'num_vars': len(params_in),
                   'names': params_in,
                   'bounds': bounds_in}

    # generates N*(2D+2) samples, where N is argument D is number of non-constant variables
    X, problem_reduced = saltelli_with_constant_bounds(problem_all, N_in)

    # If quantity_mean and quantity_variance are lists corresponding to time series, extract scalar features

    # Train and Sample Gaussian Process emulator at points
    y = gpe.train_and_predict(df_in, params_in, quantity_mean_in, quantity_variance_in, X, scalar_mean_function,
                              scalar_variance_function)

    # Input thereduced problem dictionary, the one with constants removed
    Si = sobol.analyze(problem_reduced, y)

    # make into pandas dataframe
    total_Si, first_Si, second_Si = Si.to_df()

    df_out = pd.concat([first_Si, total_Si], axis=1)
    return df_out
